#!/usr/bin/env python
#-*- coding: utf-8 -*-

"""
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
Basic script for UVEX Control.
The purpose is to provide a basic Python script to offer a demo tool.
It is to be considered as an example. This is NOT designed for production.
- Original Version : F. Cochard - May 2022
#----------------------------------------------------------------------------------
#----------------------------------------------------------------------------------
"""
#----------------------------------------------------------------------------------
# Loading libraries
#----------------------------------------------------------------------------------

import cmd # for command line instructions
from threading import Lock # to manage multi-threading
import time # gère le timing (for the function sleep)
import serial # Manages the serial port
import serial.tools.list_ports # Manages the serial ports list

#----------------------------------------------------------------------------------------------
# Variables definition
#----------------------------------------------------------------------------------------------
# constants
TIMEOUT_VALUE = 3
order_err = ['Ok, order completed', 'Invalid arguments', 'Timeout reached', 'Unexpected reply', 'Serial PORT not available', 'Checksum error in the returned message']
Locker = Lock() # Allows to lock processes during variable writing

#----------------------------------------------------------------------------------------------
# Functions definition
#----------------------------------------------------------------------------------------------

def format(str):
	"Formate une commande RS232"
	a = hex(checksum(str))[2:].zfill(2).upper() # 2: to remove \x, zfill for 2 car
	chaine = str
	return chaine

def checksum(str):
	cks = 0
	stringByte = bytes(str,'utf-8')
	for car in stringByte :
		cks = cks ^ car
	return cks

def send_order(texte):
# The function returns an error code :
    trame = format(texte)
    if (SerialPortAvailable == True) :
        try: # We check that serial port is up & running
            port_serie.reset_input_buffer() # to flush the buffer		
            port_serie.write(trame.encode('ascii')) # Sends the message
            Reply_received = False # Checks if a message is received
            timeout_limit = time.time() + TIMEOUT_VALUE
            timeout_reached = False
            while (Reply_received == False and timeout_reached == False): 
                try:
                    if port_serie.in_waiting:
                        ligne = str(port_serie.readline(),'ascii')
                        Reply_received = True 
                except:
                    print("Bug with serial port - " + trame)
                    time.sleep(3) # Waits for 3 seconds before starting again
                if (time.time() >= timeout_limit) :
                    timeout_reached = True
            if (Reply_received == True) :
                return(0, ligne, trame) # Message received OK
            if (timeout_reached == True) :
                return(2, -1, trame) # error 2 = timeout
        except serial.SerialException:
            print('The USB port is not available... anymore')
            return (4, -1, trame)
    else :
        print('The USB port is not available')
        return(4, -1, trame) # Failed operation

def send_RAW_order(text):
# same as send_order, but with no formatting!
    if (SerialPortAvailable == True) :
        try: # We check that serial port is up & running
            port_serie.reset_input_buffer() 
            port_serie.reset_output_buffer() 
            port_serie.write(text.encode('ascii')) # sends the message
            port_serie.flush() # wait until all data are transmitted
            Reply_received = False # To check if message is received
            timeout_limit = time.time() + TIMEOUT_VALUE
            timeout_reached = False
            while (Reply_received == False and timeout_reached == False):
                try:
                    if port_serie.in_waiting:
                        ligne = str(port_serie.readline(),'ascii')
                        Reply_received = True # Ok, message received
                except:
                    print("Bug with serial port - " + text)
                    time.sleep(3) # Waits for 3 seconds before starting again
                if (time.time() >= timeout_limit) :
                    timeout_reached = True
            if (Reply_received == True) :
                return(0, ligne, text) # Message received OK
            if (timeout_reached == True) :
                return(2, -1, text) # error 2 = timeout
        except serial.SerialException:
            print('The USB port is not available... anymore')
            return (4, -1, text)
    else :
        print('The USB port is not available')
        return(4, -1, text) # Failed operation

def val_message(message_reply): # to extract the value from the USIS message
    MessageContent = message_reply.split(';')
    print(MessageContent)
    value = MessageContent[4].split('\n')[0] # we remove the 'line return' (\n) at the end
    error = MessageContent[0]
    return error, value


#----------------------------------------------------------------------------------------------
# End of functions definition
#----------------------------------------------------------------------------------------------

#----------------------------------------------------------------------------------------------
# Command line class definition
#----------------------------------------------------------------------------------------------

class UVEX(cmd.Cmd):
    """ This class offers a command line tool.
    Each method starting with 'do_' is a command available for the user"""
    intro = 'Welcome to UVEX control system.\nType help or ? to list commands'
    prompt = '\n> '

    # User commands 
    #----------------------------------------------------------------------------------------------
    # 1 - Commands for the whole shelter
    #----------------------------------------------------------------------------------------------

    def do_rawMessage(self, arg):
        '''to send a RAW message (with no formatting)'''
        print("Raw message", (len(arg.split(' '))))
        if (len(arg.split(' ')) == 1): # tests if there is exactly 1 argument	
            RawText = arg.split(' ')[0] + '\n'
            print(f'Send the message: {RawText}')
            reply, val, order = send_RAW_order(RawText)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\n... Message returned: {val}')
        else :
            print('Requires exactly 1 argument (nb of blinks). Ex : test 3')


    def do_test(self, arg):
        '''test - Test the system (switches the UVEX LED on & off).'''
        pass
   
    def do_bye(self, arg):
        '''bye - to quit the shelter control program.'''
        if (SerialPortAvailable == True) :
            port_serie.close()
        print('End of the script.\nGood bye!')
        quit()

    def do_stop(self, arg):
        '''This function stops immediately stop the movement of the property, if appropriate. Format: stop property (ex: stop GRATING_ANGLE).'''
        Message = 'STOP;ALL'
        reply, val, order = send_order(Message)
        print(f'Order sent: {order}\n... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')

    def do_foc(self, arg):
        '''To get the FOCUS position (with no argument) or to move to a specific FOCUS position (argment = position in mm).'''
        if (len(arg.split()) == 0): # no argument : this is a GET request
            Message = 'GET;FOCUS_POSITION;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) == 1): # One argument : this is a SET request
            FocusTarget = arg.split()[0]
            Message = 'SET;FOCUS_POSITION;VALUE;' + str(FocusTarget) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) > 1): # There are 2 argument : we return the ATTRIBUTE attribute
            print('Requires 1 argument max (focus target)')

    def do_grat(self, arg):
        '''To get the GRATING ANGLE position (with no argument) or to move to a specific GRATING ANGLE position (argment = position in degrees).'''
        if (len(arg.split()) == 0): # no argument : this is a GET request
            Message = 'GET;GRATING_ANGLE;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) == 1): # One argument : this is a SET request
            AngleTarget = arg.split()[0]
            Message = 'SET;GRATING_ANGLE;VALUE;' + str(AngleTarget) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) > 1): # There are 2 argument : we return the ATTRIBUTE attribute
            print('Requires 1 argument max (Angle target)')

    def do_wave(self, arg):
        '''To get the current WAVELENGTH (with no argument) or to move to a specific WAVELENGTH (argment = wavelength in nm).'''
        if (len(arg.split()) == 0): # no argument : this is a GET request
            Message = 'GET;GRATING_WAVELENGTH;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) == 1): # One argument : this is a SET request
            AngleTarget = arg.split()[0]
            Message = 'SET;GRATING_WAVELENGTH;VALUE;' + str(AngleTarget) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) > 1): # There are 2 argument : we return the ATTRIBUTE attribute
            print('Requires 1 argument max (wavelength target)')

    def do_density(self, arg):
        '''To get the current grating density (nb of lines / mm - with no argument) or to change the density (argment = nb of lines / mm).'''
        if (len(arg.split()) == 0): # no argument : this is a GET request
            Message = 'GET;GRATING_DENSITY;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) == 1): # One argument : this is a SET request
            AngleTarget = arg.split()[0]
            Message = 'SET;GRATING_DENSITY;VALUE;' + str(AngleTarget) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        if (len(arg.split()) > 1): # There are 2 argument : we return the ATTRIBUTE attribute
            print('Requires 1 argument max (wavelength target)')

    def do_calgrat(self, arg):
        '''To calibrate the grating angle at current position'''
        if (len(arg.split()) == 1): # One argument is required
            CurrentAngle = arg.split()[0]
            Message = 'CALIB;GRATING_ANGLE;' + str(CurrentAngle) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires 1 argument  (current angle position)')

    def do_calfoc(self, arg):
        '''To calibrate the focus at current position'''
        if (len(arg.split()) == 1): # One argument is required
            CurrentAngle = arg.split()[0]
            Message = 'CALIB;FOCUS_POSITION;' + str(CurrentAngle) +'\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires 1 argument (focus current position)')

    def do_sky(self, arg):
        '''To switch light source to SKY position'''
        if (len(arg.split()) == 0): # One argument is required
            Message = 'SET;LIGHT_SOURCE;VALUE;SKY\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_dark(self, arg):
        '''To switch light source to DARK position'''
        if (len(arg.split()) == 0): # One argument is required
            Message = 'SET;LIGHT_SOURCE;VALUE;DARK\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_calib(self, arg):
        '''To switch light source to CALIB position'''
        if (len(arg.split()) == 0): # One argument is required
            Message = 'SET;LIGHT_SOURCE;VALUE;CALIB\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_flat(self, arg):
        '''To switch light source to FLAT position'''
        if (len(arg.split()) == 0): # One argument is required
            Message = 'SET;LIGHT_SOURCE;VALUE;FLAT\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_source(self, arg):
        '''Returns the current light source position'''
        if (len(arg.split()) == 0): # No argument is required
            Message = 'GET;LIGHT_SOURCE;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply} - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_temp(self, arg):
        '''Returns the PICO internal temperature'''
        if (len(arg.split()) == 0): # No argument is required
            Message = 'GET;TEMPERATURE;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply}  - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_version(self, arg):
        '''Returns the firmware and protocol (USIS) version'''
        if (len(arg.split()) == 0): # No argument is required
            Message = 'GET;SOFTWARE_VERSION;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply}  - {order_err[reply]}\nMessage returned: {val}')
            Message = 'GET;PROTOCOL_VERSION;VALUE\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply}  - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_reboot(self, arg):
        '''Reboots the PICO to prepare a firmware upgrade.'''
        if (len(arg.split()) == 0): # No argument is required
            Message = 'SYSTEM;REBOOT;\n'
            reply, val, order = send_order(Message)
            print(f'Order sent: {order}... reply: {reply}  - {order_err[reply]}\nMessage returned: {val}')
        else:
            print('Requires no argument')

    def do_introspection(self, arg):
        '''Lists all properties and attributes of the USIS deviceINFO.'''
        if (len(arg.split()) == 0): # No argument is required
            Message = 'GET;DEVICE_NAME;VALUE\n'
            reply, val, order = send_order(Message)
            Error, Mess = val_message(val)
            if (Error == 'M00'): # We check that the reply is OK
                print(f'USIS device name: {Mess}')
            else:
                print(f'Error message (1)...')

            Message = 'GET;SOFTWARE_VERSION;VALUE\n'
            reply, val, order = send_order(Message)
            Error, Mess = val_message(val)
            if (Error == 'M00'): # We check that the reply is OK
                print(f'USIS firmware version: {Mess}')
            else:
                print(f'Error message (2)...')

            Message = 'INFO;PROPERTY_COUNT;\n'
            reply, val, order = send_order(Message)
            Error, Mess = val_message(val)
            if (Error == 'M00'): # We check that the reply is OK
                Nb_properties = int(Mess)
                print(f'The device contains {Nb_properties} properties')
            else:
                print(f'Error message (3)...')

            for property in range(Nb_properties):
                Message = 'INFO;PROPERTY_NAME;' + str(property) + '\n'
                reply, val, order = send_order(Message)
                Error, Mess = val_message(val)
                if (Error == 'M00'): # We check that the reply is OK
                    Message = 'INFO;PROPERTY_TYPE;' + str(property) + '\n'
                    reply, val, order = send_order(Message)
                    Error1, Mess1 = val_message(val)
                    if (Error1 == 'M00'): # We check that the reply is OK
                        print(f'Property {property}: {Mess} ({Mess1})')
                        Message = 'INFO;PROPERTY_ATTR_COUNT;' + str(property) + '\n'
                        reply, val, order = send_order(Message)
                        Error2, Mess2 = val_message(val)
                        if (Error2 == 'M00'): # We check that the reply is OK
                            Nb_attributes = int(Mess2)
                            for attrib in range(Nb_attributes):
                                Message = 'INFO;PROPERTY_ATTR_NAME;' + str(property) + ';' + str(attrib) + '\n'
                                reply, val, order = send_order(Message)
                                Error3, Mess3 = val_message(val)
                                Message = 'INFO;PROPERTY_ATTR_MODE;' + str(property) + ';' + str(attrib) + '\n'
                                reply, val, order = send_order(Message)
                                Error4, Mess4 = val_message(val)
                                Message = 'GET;' + str(Mess) + ';' + str(Mess3) + '\n'
                                reply, val, order = send_order(Message)
                                Error5, Mess5 = val_message(val)
                                print(f'  > {Mess3} ({Mess4}): {Mess5}')
                                if (Mess1 == 'ENUM'):
                                    Message = 'INFO;PROPERTY_ATTR_ENUM_COUNT;' + str(property) + ';0\n'
                                    reply, val, order = send_order(Message)
                                    Error6, Mess6 = val_message(val)
                                    if (Error6 == 'M00'): # We check that the reply is OK
                                        Nb_ENUM_values = int(Mess6)
                                        MessageEnum = '  > Possible values: '
                                        for Enum in range(Nb_ENUM_values):
                                            Message = 'INFO;PROPERTY_ATTR_ENUM_VALUE;' + str(property) + ';' + str(Enum) + '\n'
                                            reply, val, order = send_order(Message)
                                            Error7, Mess7 = val_message(val)
                                            MessageEnum += Mess7 + ' - '
                                        print(f'{MessageEnum}')
                                    else:
                                        print(f'Error message (6)...')
                                    
                else:
                    print(f'Error message (4)...')
        else:
            print('Requires no argument')

#----------------------------------------------------------------------------------------------
# Main program
#----------------------------------------------------------------------------------------------

# Open the serial port (once, to avoid auto-reset)
try: # Gets the list of the USB ports
    ports = list(serial.tools.list_ports.comports())
    NbPorts = len(ports)
    print("Here is the list of the available USB ports on your computer")
    for Idx in range(0, NbPorts):
        print(f"{Idx+1} : {ports[Idx][0]}")
    Choice = int(input("Which USB port do you want to open (give the number)?\n") or "1")-1
    found_port = ports[Choice][0]
    print('Opening port :' + found_port)
    try: # To check that USB ports is available
        port_serie = serial.Serial(port=found_port, baudrate=9600, timeout=1, writeTimeout=1)
        print("The USB port is open.")
    except serial.SerialException:
        print("Unable to open the USB port.")
    SerialPortAvailable = True
except serial.SerialException:
	print('No USB port is available')
	port_serie = 0
	SerialPortAvailable = False

# Run the CMD loop (command line interpreter)
UVEX().cmdloop()

